#include <stdio.h>
#include <sys/socket.h>
#include <sys/syslog.h>

#include <bluetooth/bluetooth.h>
#include <bluetooth/hci.h>
#include <bluetooth/hci_lib.h>

#include <dbus/dbus.h>

#include "hcid.h"
#include "glib-ectomy.h"

static DBusConnection *connection;

#define TIMEOUT (30*1000)		// 30 seconds

#define REQUEST_NAME "org.handhelds.gpe.bluez.pin-request"

struct pin_request
{
  int dev;
  bdaddr_t bda;
};

static DBusHandlerResult
reply_handler_function (DBusMessageHandler *handler,
			DBusConnection     *connection,
			DBusMessage        *message,
			void               *user_data)
{
  struct pin_request *req = (struct pin_request *)user_data;
  pin_code_reply_cp pr;
  DBusMessageIter iter;
  int type;
  size_t len;
  char *pin;

  if (dbus_message_get_is_error (message))
    goto error;

  dbus_message_iter_init (message, &iter);

  type = dbus_message_iter_get_arg_type (&iter);
  if (type != DBUS_TYPE_STRING)
    goto error;

  pin = dbus_message_iter_get_string (&iter);
  len = strlen (pin);

  memset (&pr, 0, sizeof(pr));
  bacpy (&pr.bdaddr, &req->bda);
  memcpy (pr.pin_code, pin, len);
  pr.pin_len = len;
  hci_send_cmd(req->dev, OGF_LINK_CTL, OCF_PIN_CODE_REPLY,
	       PIN_CODE_REPLY_CP_SIZE, &pr);

  return DBUS_HANDLER_RESULT_REMOVE_MESSAGE;

 error:
  hci_send_cmd(req->dev, OGF_LINK_CTL, OCF_PIN_CODE_NEG_REPLY, 6, &req->bda);
  return DBUS_HANDLER_RESULT_REMOVE_MESSAGE;
}

static void
free_pin_req (void *req)
{
  free (req);
}
     
void
hcid_dbus_request_pin (int dev, struct hci_conn_info *ci)
{
  DBusMessage *message;
  DBusMessageIter iter;
  DBusMessageHandler *reply_handler;
  struct pin_request *req;
  bdaddr_t ba;
  int i;

  message = dbus_message_new (REQUEST_NAME, DBUS_SERVICE_BROADCAST);
  if (message == NULL)
    {
      syslog (LOG_ERR, "Couldn't allocate D-BUS message");
      goto failed;
    }

  req = malloc (sizeof (*req));
  req->dev = dev;
  bacpy (&req->bda, &ci->bdaddr);

  reply_handler = dbus_message_handler_new (reply_handler_function, req, free_pin_req);

  dbus_message_append_iter_init (message, &iter);

  dbus_message_iter_append_boolean (&iter, ci->out);
  for (i = 0; i < 6; i++)
    {
      unsigned char *d = (unsigned char *)&ci->bdaddr;
      dbus_message_iter_append_byte (&iter, d[i]);
    }

  baswap (&ba, &ci->bdaddr);

  if (dbus_connection_send_with_reply (connection, message, reply_handler, TIMEOUT) == FALSE)
    {
      syslog (LOG_ERR, "D-BUS send failed");
      goto failed;
    }

  dbus_connection_flush (connection);

  return;

 failed:
  hci_send_cmd(dev, OGF_LINK_CTL, OCF_PIN_CODE_NEG_REPLY, 6, &ci->bdaddr);
}

gboolean
watch_func (GIOChannel *chan, GIOCondition cond, gpointer data)
{
  int flags = 0;
  DBusWatch *watch = (DBusWatch *)data;

  if (cond & G_IO_IN) flags |= DBUS_WATCH_READABLE;
  if (cond & G_IO_OUT) flags |= DBUS_WATCH_WRITABLE;
  if (cond & G_IO_HUP) flags |= DBUS_WATCH_HANGUP;
  if (cond & G_IO_ERR) flags |= DBUS_WATCH_ERROR;

  dbus_watch_handle (watch, flags);
  
  dbus_connection_ref (connection);
  
  /* Dispatch messages */
  while (dbus_connection_dispatch (connection) == DBUS_DISPATCH_DATA_REMAINS)
    ;

  dbus_connection_unref (connection);

  return TRUE;
}

dbus_bool_t
add_watch (DBusWatch *watch, void *data)
{
  GIOCondition cond = G_IO_HUP | G_IO_ERR;
  int fd;
  GIOChannel *io;
  int flags;
  guint id;

  if (!dbus_watch_get_enabled (watch))
    return TRUE;

  fd = dbus_watch_get_fd (watch);
  io = g_io_channel_unix_new (fd);
  flags = dbus_watch_get_flags (watch);
  
  if (flags & DBUS_WATCH_READABLE) cond |= G_IO_IN;
  if (flags & DBUS_WATCH_WRITABLE) cond |= G_IO_OUT;
  
  id = g_io_add_watch (io, cond, watch_func, watch);

  dbus_watch_set_data (watch, (void *)id, NULL);

  return TRUE;
}

static void
remove_watch (DBusWatch *watch, void *data)
{
  guint id = (guint)dbus_watch_get_data (watch);

  dbus_watch_set_data (watch, NULL, NULL);

  if (id)
    g_io_remove_watch (id);
}

static void
watch_toggled (DBusWatch *watch, void *data)
{
  /* Because we just exit on OOM, enable/disable is
   * no different from add/remove
   */
  if (dbus_watch_get_enabled (watch))
    add_watch (watch, data);
  else
    remove_watch (watch, data);
}

gboolean
hcid_dbus_init (void)
{
  DBusError error;

  dbus_error_init (&error);
  connection = dbus_bus_get (DBUS_BUS_SYSTEM, &error);
  if (connection == NULL)
    {
      fprintf (stderr, "Failed to open connection to system message bus: %s\n",
               error.message);
      dbus_error_free (&error);
      return FALSE;
    }

  dbus_connection_set_watch_functions (connection, add_watch, remove_watch,
				       watch_toggled, NULL, NULL);

  return TRUE;
}
